with Ada.Strings.Text_Buffers;

generic
   Capacity : Positive;
   type Element_T is private;
package Generic_Stacks is
   type Stack is private with Put_Image => Put_Image;

   function Empty (S : Stack) return Boolean;
   function Full (S : Stack) return Boolean;

   procedure Put_Image
     (Output : in out Ada.Strings.Text_Buffers.Root_Buffer_Type'Class;
      S      : Stack);

   type Stack_Model is array (Positive range <>) of Element_T with Ghost;
   function Model (S : Stack) return Stack_Model
   with Ghost;
   function Pop (S : in out Stack) return Element_T
   with Pre => not Empty (S), Post => Model (S) & Pop'Result = Model (S)'Old;
   procedure Push (S : in out Stack; E : Element_T)
   with Pre => not Full (S), Post => Model (S) = Model (S)'Old & E;
private
   subtype Stack_Index is Natural range 1 .. Capacity;
   subtype Stack_Count is Natural range 0 .. Stack_Index'Last;
   type Stack_Data is array (Stack_Index) of Element_T;
   type Stack is record
      Count : Stack_Count := 0;
      Data  : Stack_Data;
   end record;

   function Empty (S : Stack) return Boolean
   is (S.Count = 0);
   function Full (S : Stack) return Boolean
   is (S.Count = Stack_Count'Last);
end Generic_Stacks;
